//this file is part of eMule
//Copyright (C)2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#pragma once
#include "EMSocket.h"

class CUpDownClient;
class CPacket;
class CTimerWnd;

enum SocketState 
{
	SS_Other,		//These are sockets we created that may or may not be used.. Or incoming connections.
	SS_Half,		//These are sockets that we called ->connect(..) and waiting for some kind of response.
	SS_Complete	//These are sockets that have responded with either a connection or error.
};

class CClientReqSocket : public CEMSocket
{
	friend class CListenSocket;
	DECLARE_DYNCREATE(CClientReqSocket)

public:
	CClientReqSocket(CUpDownClient* in_client = NULL);	
	void	SetClient(CUpDownClient* pClient);
	void	Disconnect(LPCTSTR pszReason);
	void	WaitForOnConnect();
	// Pawcio:BC - moved
	bool	CheckTimeOut();
	virtual UINT GetTimeOut();
	virtual void Safe_Delete();
	
	bool	Create();
	virtual void SendPacket(Packet* packet, bool delpacket = true, bool controlpacket = true, uint32 actualPayloadSize = 0);
    virtual SocketSentBytes Send(uint32 maxNumberOfBytesToSend); // Pawcio: BC

	void	DbgAppendClientInfo(CString& str);
	CString DbgGetClientInfo();

	CUpDownClient*	client;
	// Pawcio: BC
	void		ResetSendTimes() {m_dwLastBlockKeepAliveSent = m_dwLastBlockSent = ::GetTickCount();}
	void		ResetProcessRate();
	void		SetLastBlockSent() {m_dwLastBlockSent=::GetTickCount();}
	uint8		processed;
	uint32		m_uBusyTicks;
	uint32		m_uReadyTicks;
	float		GetAvgRatio()		{return avg_ratio;}
	float		GetCurRatio()		{return cur_ratio;}
	void		CalcRatio(bool busy = false);
	void		SetTimeForSend(uint32 time, bool keepalive) { if (!keepalive) m_dwLastBlockSent = time; m_dwLastBlockKeepAliveSent = time;}
	uint32		GetTimeSinceLastSent()		{ return ::GetTickCount() - m_dwLastBlockSent;}
	uint32		GetTimeSinceLastKeepAliveSent()		{ return ::GetTickCount() - m_dwLastBlockKeepAliveSent;}
	void		ConsiderReservedTime() {m_dwLastBlockSent=m_dwLastBlockKeepAliveSent;}
	bool		isForPriorityClient;
	bool		onReserved;
	// <--- Pawcio: BC
protected:
	virtual ~CClientReqSocket();
	virtual void Close() { CAsyncSocketEx::Close(); }
	void	Delete_Timed();

	virtual void OnConnect(int nErrorCode);
	void		 OnClose(int nErrorCode);
	void		 OnSend(int nErrorCode);
	void		 OnReceive(int nErrorCode);
	void		 OnError(int nErrorCode);
	virtual bool PacketReceived(Packet* packet);
	int			 PacketReceivedSEH(Packet* packet);
	bool		 PacketReceivedCppEH(Packet* packet);

	bool	ProcessPacket(const BYTE* packet, uint32 size,UINT opcode);
	bool	ProcessExtPacket(const BYTE* packet, uint32 size, UINT opcode, UINT uRawSize);
#ifdef WEBCACHE
	bool	ProcessWebCachePacket(const BYTE* packet, uint32 size, UINT opcode, UINT uRawSize); // yonatan - webcache protocol packets
#endif
	void	PacketToDebugLogLine(LPCTSTR protocol, const uchar* packet, uint32 size, UINT opcode, EDebugLogPriority dlpPriority);
	void	SetConState(SocketState val);

	// Pawcio: BC - moved
	bool	deletethis;
	uint32	deltimer;
	bool	m_bPortTestCon;
	uint32	m_nOnConnect;
	// Pawcio: BC
	float		cur_ratio;
	float		avg_ratio;
#define		RATIO_AVG	100
	float		ratios[RATIO_AVG];
	uint8		ratioIndex;
	uint32		m_dwLastBlockSent;
	uint32		m_dwLastBlockKeepAliveSent;
	// <--- Pawcio: BC
};


class CListenSocket : public CAsyncSocketEx
{
	friend class CClientReqSocket;
public:
	CListenSocket();
	virtual ~CListenSocket();

	bool	StartListening();
	void	StopListening();
	virtual void OnAccept(int nErrorCode);
	void	Process();
	void	RemoveSocket(CClientReqSocket* todel);
	void	AddSocket(CClientReqSocket* toadd);
	uint16	GetOpenSockets()		{return socket_list.GetCount();}
	void	KillAllSockets();
	bool	TooManySockets(bool bIgnoreInterval = false);
	uint32	GetMaxConnectionReached()	{return maxconnectionreached;}
	bool    IsValidSocket(CClientReqSocket* totest);
	void	AddConnection(bool created); // Pawcio: BC
	void	RecalculateStats();
	void	ReStartListening();
	void	Debug_ClientDeleted(CUpDownClient* deleted);
	bool	Rebind();
	bool	SendPortTestReply(char result,bool disconnect=false);

	void	UpdateConnectionsStatus();
	float	GetMaxConperFiveModifier();
	uint32	GetPeakConnections()		{ return peakconnections; }
	uint32	GetTotalConnectionChecks()	{ return totalconnectionchecks; }
	float	GetAverageConnections()		{ return averageconnections; }
	uint32	GetActiveConnections()		{ return activeconnections; }
	uint16	GetConnectedPort()			{ return m_port; }
	uint32	GetTotalHalfCon()			{ return m_nHalfOpen; }
	uint32	GetTotalComp()				{ return m_nComp; }

	void	SetLimitForConnections(); //Pawcio: BC
	void	ApplyToLimits(bool created); //Pawcio: BC
private:
	bool bListening;
	CTypedPtrList<CPtrList, CClientReqSocket*> socket_list;
	uint16	opensockets;
	uint16	m_OpenSocketsInterval;
	uint32	maxconnectionreached;
	uint16	m_ConnectionStates[3];
	int		m_nPendingConnections;
	uint32	peakconnections;
	uint32	totalconnectionchecks;
	float	averageconnections;
	uint32	activeconnections;
	uint16  m_port;
	uint32	m_nHalfOpen;
	uint32	m_nComp;
	//Pawcio: BC
	int		m_CreateSocketsLimit;	
	int		m_AcceptSocketsLimit;	
	int		m_AllSocketsLimit;	
	//End Pawcio
};
